/**********************************************************************/
/*                                                                    */ 
/* File   : ChatServer.java                                           */
/* Author : WEI-NING LIN                                              */
/*                                                                    */
/* description :                                        */
/*  keeps track of the active clients and updates clients with the  */
/*  new messages and the new user list.                               */
/*  Handles each client's request to broadcast the encrypted messages  */ 
/*  to other clients
/**********************************************************************/

import java.rmi.*;
import java.awt.Point;
import java.rmi.server.*;
import java.rmi.registry.LocateRegistry;
import java.util.*;


public class ChatServer extends UnicastRemoteObject
    implements ChatServerInterface
{


    //maintaines the master copy of user list and public key list
    //clients update their user list and public key list from the 
    //server
    private Hashtable broadcastList = new Hashtable();
    private Hashtable broadcastKeyList = new Hashtable();

    ChatAppletInterface chatClient;
    String user;

    public ChatServer() throws RemoteException
    {
        super(); // invokes the constructor of
                 // java.rmi.server.UnicastRemoteObject, which export
                 // the remote object by listening for incoming calls
                 // to the remote object on an anonymous port.
    }


    //server, not an applet thus has a main
    public static void main(String args[])
    {
        // a security manager needs to be running so that it can
        // guarantee that the classes loaded do not perform 
        // sensitive operatons. if not, no class loading for RMI
        // classes.
        System.setSecurityManager(new RMISecurityManager());

        try
        {
            // create the registry.
            System.out.println("ChatServer.main: creating registry");
            LocateRegistry.createRegistry(2002);

            // create the server.
            System.out.println("ChatServer.main: creating server");
            ChatServer chatServer = new ChatServer();

            // bind the server to the registry
            System.out.println("ChatServer.main: binding server ");

            Naming.rebind("//carver.npac.syr.edu:2002/ChatServer",
            chatServer);
            System.out.println("ChatServer.main: Done");
        }
        catch (Exception e)
        {
            System.out.println("ChatServer error: " + e.getMessage());
            e.printStackTrace();
        }
    }

    // called by client. Updates  the newly joined client's user list
    public void connect(ChatAppletInterface client) throws RemoteException
    {
        if (!broadcastList.isEmpty())
        {
            Vector v = getUserList();
            Vector v1 = getUserKey();

            // tell the newly connected client to update it's user list.
            try
            {
                client.updateUserList(v,v1);
            }
            catch (Exception e)
            {
                System.out.println("ChatServer error: " + e.getMessage());
                e.printStackTrace();
            }
        }
    }

    //called by client when loggin off
    public void logoff(String user) throws RemoteException
    {
        // remove the client from the broadcast list
        // and let everyone else know client has left the conversation


        Vector userList = getUserList();
        Vector keyList = getUserKey();

        // enumerat connected clients.
        userList.removeElement(user);
        Enumeration e = broadcastList.elements();

        //updates clients' user list
        while (e.hasMoreElements())
        {
           //updates every client's user and pub key list
            ChatAppletInterface c = (ChatAppletInterface) e.nextElement();
            updateUserList(c, userList, keyList);
            c.displayMessage(user + " leaves the conversation");
        }
        broadcastList.remove(user);
        broadcastKeyList.remove(user);
    }

    //called when a client logins
    public void login(ChatAppletInterface client, String user, Point pk) throws RemoteException
    {
        // if the user is not in the broadcast list the add the
        // user as the key and the client as the value.
        if (!broadcastList.containsKey(user))
        {
            broadcastList.put(user, client);
            broadcastKeyList.put(user, pk);
        }

        Vector userList = getUserList();
        Vector keyList = getUserKey();

        // enumerat connected clients.
        Enumeration e = broadcastList.elements();
        while (e.hasMoreElements())
        {

           //updates every client's user and pub key list
            ChatAppletInterface c = (ChatAppletInterface) e.nextElement();
            updateUserList(c, userList, keyList);
            c.displayMessage(user + " has entered the conversation\n");
        }
    }

    public void mouseDrag(String ruser,long x, long y) throws RemoteException
    {
    
        ChatAppletInterface client =
               (ChatAppletInterface) broadcastList.get(ruser);
           
        client.mouseDrag(x, y);
    }
    public void mouseDown(String ruser, long x, long y) throws RemoteException
    {
        ChatAppletInterface client =
               (ChatAppletInterface) broadcastList.get(ruser);
           
        client.mouseDown(x, y);

    }
    public void broadcastMessage(String user, String ruser, long emsg[]) throws RemoteException
    {
        _broadcastMessage(user, ruser, emsg);
    }

    //broadcasts the encrpted message sent by the client
    public void _broadcastMessage(String user, String ruser, long emsg[])
    {
            ChatAppletInterface client =
                (ChatAppletInterface) broadcastList.get(ruser);

            try
            {
                StringBuffer sb = new StringBuffer(user);
                sb.append(" said > ");
                client.displayMessage(sb.toString() );

             //sends the encrypted message
                client.displayMessage(emsg);
            }
            catch (Exception e)
            {
                System.out.println("ChatServer error: " + e.getMessage());
                e.printStackTrace();
            }
    }

    //extracts the list of clients' public keys
    private Vector getUserKey()
    {
        Vector v = new Vector();

        // make a Vector of users.
        Enumeration e = broadcastKeyList.elements();
        while (e.hasMoreElements())
        {
            Point user = (Point) e.nextElement();
            v.addElement(user);
        }

        return v;
    }

    //gets the active user list
    private Vector getUserList()
    {
        Vector v = new Vector();

        // make a Vector of users.
        Enumeration e = broadcastList.keys();
        while (e.hasMoreElements())
        {
            String user = (String) e.nextElement();
            v.addElement(user);
        }

        return v;
    }

    // update client's user and public key list.
    private void updateUserList(ChatAppletInterface client, Vector userList,
                        Vector keyList)
    {
        try
        {
            client.updateUserList(userList, keyList);
        }
        catch (Exception e)
        {
            System.out.println("ChatServer error: " + e.getMessage());
            e.printStackTrace();
        }
    }
}
